
////////////////////////////////////////////////////////////////////
import { vec3, vec4 } from "gl-matrix";
const hasPassive = require('has-passive-events');
import { OBJ } from 'webgl-obj-loader';

// wgtek ///////////////////////////////////////////////////////////
import { wgPrimRender } from "wgTek/wgPrimRender.js"
import { wgContext } from "wgTek/wgContext.js"
import { wgShader } from "wgTek/wgShader.js"
import { wgVertexBuffer, wgIndexType } from "wgTek/wgVertexBuffer.js"
import { VertexType } from "wgTek/wgVertexDescriptors.js"
import { wgVertexDescriptor } from "wgTek/wgVertexDescriptors.js"

// wfundament /////////////////////////////////////////////////////
import { MouseEventHandler } from "wFundament/MouseEventHandler.js"
import { KeyboardEventHandler } from "wFundament/KeyboardEventHandler.js"
import { TouchEventHandler } from "wFundament/TouchEventHandler.js"
import { wOrbitCamera } from "wFundament/wOrbitCamera.js"
import { Vec4Colors } from "wFundament/wContants";
import { wBlockLetters } from "./BlockLetters.js"
////////////////////////////////////////////////////////////////////



// Snake game logic ////////////////////////////////////////////////
const SnakeHeadings = {
    West: [1, 0],
    East: [-1, 0],
    North: [0, 1],
    South: [0, -1],
}

const EmptyValue = 0;
const WallValue = 1;
const SnakeValue = 2;
const FoodValue = 3;

class Snake {
    constructor() {
        this.heading_ = SnakeHeadings.West;
        this.indices_ = [];
        this.resetSnake();
    }

    getSnakeHead() {
        return this.indices_[0];
    }

    getSnakeTail() {
        return this.indices_[this.indices_.length - 1];
    }

    resetSnake() {
        this.indices_.length = 0;
        this.heading_ = SnakeHeadings.West;
        this.indices_.push([8, 5]);
        this.indices_.push([7, 5]);
        this.indices_.push([6, 5]);
        this.indices_.push([5, 5]);
    }
}

class SnakeGameLogic {

    constructor(dim) {
        this.dimension_ = dim;
        this.level_ = new Int32Array(this.dimension_ * this.dimension_);
        this.snake_ = new Snake();
        this.resetLevel();
        this.updateTick_ = 0.0;
        this.tickTime_ = 0.3;
        this.score_ = 0;
        this.gameString_ = "Press 'S' to start.";
        this.isRunning_ = false;
    }

    startGame() {
        this.snake_.resetSnake();
        this.resetLevel();
        this.isRunning_ = true;
        this.tickTime_ = 0.3;
        this.gameString_ = "score: " + this.score_;
    }

    getGameString() {
        return this.gameString_;
    }

    setSnakeHeading(heading) {
        this.snake_.heading_ = heading;
    }

    getSnakeHeading() {
        return this.snake_.heading_;
    }

    getLevelTile(x, y) {
        return this.level_[x + y * this.dimension_];
    }

    setLevelTile(x, y, value) {
        this.level_[x + y * this.dimension_] = value;
    }

    resetLevel() {
        this.score_ = 0;
        for (let y = 0; y < this.dimension_; y++) {
            for (let x = 0; x < this.dimension_; x++) {

                this.level_[x + y * this.dimension_] = EmptyValue;

                if (x == 0 || x == this.dimension_ - 1) {
                    this.level_[x + y * this.dimension_] = WallValue;
                }

                if (y == 0 || y == this.dimension_ - 1) {
                    this.level_[x + y * this.dimension_] = WallValue;
                }
            }
        }

        this.setLevelTile(21, 4, FoodValue);

        for (let i = 0; i < this.snake_.indices_.length; i++) {
            const index = this.snake_.indices_[i];
            this.setLevelTile(index[0], index[1], SnakeValue);
        }

        this.updateTick_ = 0.0;
    }

    gameOverDude() {
        this.gameString_ = "Press 'S' start.";
        this.isRunning_ = false;
    }

    updateLevel(deltaTime) {

        this.updateTick_ += deltaTime;
        if (!this.isRunning_ || this.updateTick_ < this.tickTime_) return;

        this.tickTime_ -= 0.0005;
        this.tickTime_ = Math.max(0.125, this.tickTime_);

        this.updateTick_ = 0.0;
        let snakeHead = this.snake_.getSnakeHead();

        // find the next snake position
        const nextPosX = snakeHead[0] + this.snake_.heading_[0];
        const nextPosY = snakeHead[1] + this.snake_.heading_[1];
        const headPosValue = this.getLevelTile(nextPosX, nextPosY);

        // test self intersection
        if (headPosValue == SnakeValue) {
            this.gameOverDude();
            return;
        }

        // test wall intersection
        if (headPosValue == WallValue) {
            this.gameOverDude();
            return;
        }

        // test food intersection
        let foundFood = false;
        if (headPosValue == FoodValue) {
            this.score_++;
            this.gameString_ = "score: " + this.score_;
            foundFood = true;
        }

        if (foundFood) // the snake grows when fed
        {
            while (true) {
                let rX = Math.floor(Math.random() * this.dimension_ - 1.0);
                let rY = Math.floor(Math.random() * this.dimension_ - 1.0);

                rX = Math.min(Math.max(2, rX), this.dimension_ - 2);
                rY = Math.min(Math.max(2, rY), this.dimension_ - 2);

                if (this.getLevelTile(rX, rY) == EmptyValue) {
                    this.setLevelTile(rX, rY, FoodValue);
                    break;
                }
            }
        }

        // reset previus snake tiles
        for (let i = 0; i < this.snake_.indices_.length; i++) {
            var index = this.snake_.indices_[i];
            this.setLevelTile(index[0], index[1], EmptyValue);
        }

        for (let i = this.snake_.indices_.length - 1; i > 0; i--) {
            // reset the tiles
            var index0 = this.snake_.indices_[i];
            const index1 = this.snake_.indices_[i - 1];
            index0[0] = index1[0];
            index0[1] = index1[1];
        }

        // update the head
        this.snake_.indices_[0][0] = nextPosX;
        this.snake_.indices_[0][1] = nextPosY;

        if (foundFood) // the snake grows
        {
            let tail = this.snake_.getSnakeTail();
            this.snake_.indices_.push([tail[0], tail[1]]);
        }

        // update the tile with the snake position
        for (let i = 0; i < this.snake_.indices_.length; i++) {
            const index = this.snake_.indices_[i];
            this.setLevelTile(index[0], index[1], SnakeValue);
        }

    }
}

// app /////////////////////////////////////////////////////////////
class wgApp {
    constructor(canvas) {

        // Create web gl context
        const gl = canvas.getContext('webgl2');
        if (gl === undefined) {
            window.alert("No webgl 2.0 support");
            return;
        }
        this.glContext_ = new wgContext(gl, canvas);
        this.gl = gl;

        // Create the orbit camera
        this.orbitCamera_ = new wOrbitCamera("OrbitCamera");
        this.orbitCamera_.setViewport(this.glContext_.getCanvasWidth(), this.glContext_.getCanvasHeight());
        this.orbitCamera_.setPerspective(60.0);
        this.orbitCamera_.setCameraPosition([0.0, 25.0, 10.0]);

        // Primitive render
        this.primRender_ = new wgPrimRender("PrimitiveRender", this.glContext_);
        this.blockLetters_ = new wBlockLetters();

        // Event handlers
        this.score_ = 0;
        this.numberOfElementsSpawned_ = 0;
        this.keyboardEventHandler_ = new KeyboardEventHandler();
        this.mouseEventHander_ = new MouseEventHandler();
        this.touchEventHandler_ = new TouchEventHandler();
        this.setupEventHandlers(canvas);

        // create a simple shader
        this.createShader();

        // Game elements
        this.snakeGame_ = new SnakeGameLogic(25);

        this.init().then(() => {
            this.lastTimeStamp_ = 0;
            requestAnimationFrame((timestamp) => { this.loop(timestamp); })
        });

    }

    async init() {

    }

    getKeyboardEventHandler() { return this.keyboardEventHandler_; }
    getMouseEventHandler() { return this.mouseEventHander_; }
    getTouchEventHandler() { return this.touchEventHandler_; }
    setupEventHandlers(canvas) {

        // Keyboard
        window.addEventListener("keydown", (event) => { this.getKeyboardEventHandler().keyDown(event.key) });
        window.addEventListener("keyup", (event) => { this.getKeyboardEventHandler().keyUp(event.key) });

        // Mouse
        canvas.addEventListener("mousedown", (event) => { this.getMouseEventHandler().onMouseDown(event) });
        canvas.addEventListener("mouseup", (event) => { this.getMouseEventHandler().onMouseUp(event) });
        canvas.addEventListener("mousemove", (event) => { this.getMouseEventHandler().onMouseMove(event) });
        canvas.addEventListener("mouseleave", (event) => { this.getMouseEventHandler().onMouseLeave(event) });
        canvas.addEventListener("click", (event) => { this.getMouseEventHandler().onMouseClick(event) });
        canvas.addEventListener("dblclick", (event) => { this.getMouseEventHandler().onMouseDoubleClick(event) });
        canvas.addEventListener("wheel", (event) => {
            event.preventDefault();
            this.getMouseEventHandler().onMouseWheel(event)
        });

        // Touch
        canvas.addEventListener("touchstart", (event) => { this.getTouchEventHandler().onTouchStart(event) }, hasPassive ? { passive: false } : false);
        canvas.addEventListener("touchend", (event) => { this.getTouchEventHandler().onTouchEnd(event) }, hasPassive ? { passive: false } : false);
        canvas.addEventListener("touchcancel", (event) => { this.getTouchEventHandler().onTouchCancel(event) }, hasPassive ? { passive: false } : false);
        canvas.addEventListener("touchmove", (event) => { this.getTouchEventHandler().onTouchMove(event) }, hasPassive ? { passive: false } : false);
        canvas.addEventListener('contextmenu', (e) => {
            e.preventDefault();
        }, false);
    }

    loop(timestamp) {
        let deltaTime = timestamp - this.lastTimeStamp_;
        this.update(deltaTime / 1000.0); // in milies
        this.render(timestamp);
        this.lastTimeStamp_ = timestamp;
        requestAnimationFrame((ts) => { this.loop(ts); })
    }

    update(deltaTime) {
        this.orbitCamera_.update(deltaTime * 10.0, this.getMouseEventHandler(), this.getKeyboardEventHandler(), this.getTouchEventHandler());
        this.snakeGame_.updateLevel(deltaTime);

        if (this.getKeyboardEventHandler().isDown("ArrowLeft")) {
            if (this.snakeGame_.getSnakeHeading() != SnakeHeadings.West)
                this.snakeGame_.setSnakeHeading(SnakeHeadings.East);
        }

        if (this.getKeyboardEventHandler().isDown("ArrowRight")) {
            if (this.snakeGame_.getSnakeHeading() != SnakeHeadings.East)
                this.snakeGame_.setSnakeHeading(SnakeHeadings.West);
        }

        if (this.getKeyboardEventHandler().isDown("ArrowUp")) {
            if (this.snakeGame_.getSnakeHeading() != SnakeHeadings.North)
                this.snakeGame_.setSnakeHeading(SnakeHeadings.South);
        }

        if (this.getKeyboardEventHandler().isDown("ArrowDown")) {
            if (this.snakeGame_.getSnakeHeading() != SnakeHeadings.South)
                this.snakeGame_.setSnakeHeading(SnakeHeadings.North);
        }

        if (this.getKeyboardEventHandler().isPressed('s')) {
            this.snakeGame_.startGame();
        }

        this.getMouseEventHandler().flush();
        this.getTouchEventHandler().flush();
    }

    updateGameRendering() {
        
        const dim = this.snakeGame_.dimension_;

        // Basic grid rendering 
        this.primRender_.addGridXZ(dim + 2, 1.0); 

        // offset the game to center around origo
        const xOffset = dim * 0.5;
        const zOffset = dim * 0.5;

        // Render the game environment.
        for (let y = 0; y < dim; y++) {
            for (let x = 0; x < dim; x++) {
                let gameTile = this.snakeGame_.getLevelTile(x, y);

                if (gameTile == WallValue) {
                    this.primRender_.addWireBox(vec3.fromValues(x + 0.5 - xOffset, 0.55, y + 0.5 - zOffset), vec3.fromValues(0.85, 0.85, 0.85), Vec4Colors.LightGrey);
                }

                if (gameTile == FoodValue) {
                    this.primRender_.addWireBox(vec3.fromValues(x + 0.5 - xOffset, 0.55, y + 0.5 - zOffset), vec3.fromValues(0.85, 0.85, 0.85), Vec4Colors.Yellow);
                }
            }
        }

        // Render the snake using a bunch of spheres 
        for (let i = 0; i < this.snakeGame_.snake_.indices_.length; i++) {
            let idx = this.snakeGame_.snake_.indices_[i];
            let pos = vec3.fromValues(idx[0] + 0.5 - xOffset, 0.55, idx[1] + 0.5 - zOffset);
            this.primRender_.addWireGeoSphere(pos, 0.9, Vec4Colors.White, 1);

            if (i < this.snakeGame_.snake_.indices_.length - 1) {
                let idx1 = this.snakeGame_.snake_.indices_[i + 1];
                let pos1 = vec3.fromValues(idx1[0] + 0.5 - xOffset, 0.55, idx1[1] + 0.5 - zOffset);

                let mid = vec3.fromValues(
                    pos[0] + (pos1[0] - pos[0]) * 0.5,
                    pos[1] + (pos1[1] - pos[1]) * 0.5,
                    pos[2] + (pos1[2] - pos[2]) * 0.5);

                this.primRender_.addWireGeoSphere(mid, 0.7, Vec4Colors.Red, 0);
            }
        }

        // Render the game string that contains score and information.
        let str = this.snakeGame_.getGameString();
        let iPos = -str.length * 2;
        let scale = 0.4;

        for (let i = 0; i < str.length; i++) {
            let Letter_ = this.blockLetters_.getLetter(str[i]);
            if (Letter_ === null) continue;
            for (let y = 0; y < this.blockLetters_.getHeight(); y++) {
                for (let x = 0; x < this.blockLetters_.getWidth(); x++) {
                    if (Letter_.getPixel(x, y) == 1) {
                        if (this.snakeGame_.isRunning_) {
                            this.primRender_.addWireBox(vec3.fromValues((x + iPos) * scale, (y + 5.0) * scale, -15.0), vec3.fromValues(0.9 * scale, 0.9 * scale, 0.9 * scale), Vec4Colors.Yellow);
                        }
                        else {
                            this.primRender_.addWireBox(vec3.fromValues((x + iPos) * scale, (y + 5.0) * scale, -15.0), vec3.fromValues(0.9 * scale, 0.9 * scale, 0.9 * scale), Vec4Colors.Orange);
                        }
                    }
                }
            }
            iPos += Letter_.getLetterWidth();
        }
    }

    render() {
        
        const gl = this.gl;
        this.primRender_.clearPrimitive(); 
        this.updateGameRendering();

        gl.clearColor(0.3, 0.3, 0.3, 1.0);
        gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
        gl.enable(gl.DEPTH_TEST);

        const vp = this.orbitCamera_.getViewProjection();
        this.primRender_.render(vp);
    }

    resize(width, height) {
        this.glContext_.canvas_.width = width;
        this.glContext_.canvas_.height = height;
        this.orbitCamera_.setViewport(this.glContext_.getCanvasWidth(), this.glContext_.getCanvasHeight());
        this.orbitCamera_.setPerspective(60.0);
        this.glContext_.gl_.viewport(0, 0, width, height);
    }
}

export { wgApp };
